/*****************************************************************************

This source code is made available to you under the terms of the Open
Development License.  This Package should be accompanied by a file that
contains the associated licensing terms.  If you are unable to find this file,
please refer to its location on the Internet at www.rocklyte.com/license.html.

******************************************************************************

###CLASS###
Name:      Arrow
Version:   1.1
ID:        ID_ARROW
Category:  GUI
Date:      June 2003
Author:    Rocklyte Systems
Copyright: Rocklyte Systems (c) 2000-2003.  All rights reserved.
Keywords:  arrow, pointer, point, direction, up, down, left, right
Short:     Draws arrows in various directions.

###DESCRIPTION###
<p>The Arrow class provides graphical support for drawing arrows in any
direction.  Creating an Arrow object is straight-forward, and only requires
that you set basic attributes such as the dimensions and the direction that
the Arrow should point to.  Support for setting the colour of the arrow
and other common GUI constructs is also provided.</p>

<p>The following example creates an Arrow that points to the left:</p>

<pre>
  &lt;arrow x="3" y="6" width="4" height="4" direction="left"
    colour="210,210,255"/&gt;
</pre>

<p>The Arrow class supports all common GUI constructs including offsets
and relative dimensions.  For more detailed information, refer to the
documentation for the individual fields.</p>

###END###

*****************************************************************************/

#include <pandora/main.h>
#include <pandora/system/all.h>
#include <pandora/graphics/all.h>
#include <pandora/modules/strings.h>

static ERROR CMDInit(OBJECTPTR, struct KernelBase *);
static ERROR CMDExpunge(void);

#define VER_ARROW 1.1

MODULE_HEADER = {
   MODULE_HEADER_V1, CMDInit, NULL, NULL, CMDExpunge,
   JMP_DEFAULT, 0, CPU_PC, VER_ARROW, VER_KERNEL,
   "Rocklyte Systems",
   "Rocklyte Systems (c) 2000-2004. All rights reserved.",
   "May 2004",
   "Arrow"
};

static struct StringsBase *StringsBase = 0;
static struct KernelBase *KernelBase = 0;
static OBJECTPTR ArrowClass = 0;
static OBJECTPTR StringsMod = 0;

/****************************************************************************
** Arrow class.
*/

struct Arrow {
   OBJECT_HEADER
   OBJECTID DrawableID;     /* The object that should be rendered to */
   LONG  Fill;
   LONG  Direction;
   FLOAT Scale;
   LONG  Frame;
   LONG  AlignFlags;
   LONG  ArrowWidth;
   LONG  ArrowHeight;
   LONG  ArrowX;
   LONG  ArrowY;

   /*** Virtual data store ***/
   struct RGB Highlight;
   struct RGB Shadow;
   struct RGB Colour;
   LONG  Thickness;
   LONG  Dimensions;
   LONG  Flags;
   FLOAT Width;
   FLOAT Height;
   FLOAT XCoord;
   FLOAT YCoord;
   FLOAT XOffset;
   FLOAT YOffset;
   WORD Opacity;
};

static struct FieldDef AlignFlags[] = {
   { "RIGHT",    ALIGN_RIGHT    }, { "LEFT",       ALIGN_LEFT    },
   { "BOTTOM",   ALIGN_BOTTOM   }, { "TOP",        ALIGN_TOP     },
   { "CENTER",   ALIGN_CENTER   }, { "MIDDLE",     ALIGN_MIDDLE  },
   { "VERTICAL", ALIGN_VERTICAL }, { "HORIZONTAL", ALIGN_HORIZONTAL },
   { NULL, NULL }
};

static ERROR GET_ColourRGB(struct Arrow *, struct RGB *);
static ERROR GET_HighlightRGB(struct Arrow *, struct RGB *);
static ERROR GET_Opacity(struct Arrow *, FLOAT *);
static ERROR GET_Raised(struct Arrow *, LONG *);
static ERROR GET_ShadowRGB(struct Arrow *, struct RGB *);
static ERROR GET_Sunken(struct Arrow *, LONG *);
static ERROR GET_Height(struct Arrow *, struct Variable *);
static ERROR GET_Width(struct Arrow *, struct Variable *);
static ERROR GET_X(struct Arrow *, struct Variable *);
static ERROR GET_XOffset(struct Arrow *, struct Variable *);
static ERROR GET_Y(struct Arrow *, struct Variable *);
static ERROR GET_YOffset(struct Arrow *, struct Variable *);

static ERROR SET_Colour(struct Arrow *, STRING);
static ERROR SET_ColourRGB(struct Arrow *, struct RGB *);
static ERROR SET_Highlight(struct Arrow *, STRING);
static ERROR SET_HighlightRGB(struct Arrow *, struct RGB *);
static ERROR SET_Opacity(struct Arrow *, FLOAT *);
static ERROR SET_Raised(struct Arrow *, LONG *);
static ERROR SET_Scale(struct Arrow *, FLOAT *);
static ERROR SET_Shadow(struct Arrow *, STRING);
static ERROR SET_ShadowRGB(struct Arrow *, struct RGB *);
static ERROR SET_Sunken(struct Arrow *, LONG *);
static ERROR SET_Height(struct Arrow *, struct Variable *);
static ERROR SET_Width(struct Arrow *, struct Variable *);
static ERROR SET_X(struct Arrow *, struct Variable *);
static ERROR SET_XOffset(struct Arrow *, struct Variable *);
static ERROR SET_Y(struct Arrow *, struct Variable *);
static ERROR SET_YOffset(struct Arrow *, struct Variable *);

static struct FieldDef DirectionTable[] = {
   { "UP",    ARROW_UP },
   { "DOWN",  ARROW_DOWN },
   { "LEFT",  ARROW_LEFT },
   { "RIGHT", ARROW_RIGHT },
   { NULL, NULL }
};

/*** Internal Flags ***/

#define AF_VISIBLE 0x00000001
#define AF_RAISED  0x00000002
#define AF_SUNKEN  0x00000004

static struct FieldArray ArrowFields[] = {
   { "Drawable",     0, FDF_OBJECTID|FDF_RI,        ID_RENDER,         NULL, NULL },
   { "Fill",         0, FDF_LONG|FDF_RW,            0,                 NULL, NULL },
   { "Direction",    0, FDF_LONG|FDF_LOOKUP|FDF_RW, (LONG)&DirectionTable, NULL, NULL },
   { "Scale",        0, FDF_FLOAT|FDF_RW,           0,                 NULL, SET_Scale },
   { "Frame",        0, FDF_LONG|FDF_RW,            0,                 NULL, NULL },
   { "Align",        0, FDF_LONGFLAGS|FDF_RW,       (LONG)&AlignFlags, NULL, NULL },
   { "ArrowWidth",   0, FDF_LONG|FDF_RI,            0,                 NULL, NULL },
   { "ArrowHeight",  0, FDF_LONG|FDF_RI,            0,                 NULL, NULL },
   { "ArrowX",       0, FDF_LONG|FDF_RI,            0,                 NULL, NULL },
   { "ArrowY",       0, FDF_LONG|FDF_RI,            0,                 NULL, NULL },
   /*** Virtual Fields ***/
   { "Colour",       0, FDF_STRING|FDF_W,   0, NULL,             SET_Colour },
   { "ColourRGB",    0, FDF_POINTER|FDF_RW, 0, GET_ColourRGB,    SET_ColourRGB },
   { "Highlight",    0, FDF_STRING|FDF_W,   0, NULL,             SET_Highlight },
   { "HighlightRGB", 0, FDF_POINTER|FDF_R,  0, GET_HighlightRGB, SET_HighlightRGB },
   { "Opacity",      0, FDF_FLOAT|FDF_RW,   0, GET_Opacity,      SET_Opacity },
   { "Raised",       0, FDF_LONG|FDF_RW,    0, GET_Raised,       SET_Raised },
   { "Shadow",       0, FDF_STRING|FDF_W,   0, NULL,             SET_Shadow },
   { "ShadowRGB",    0, FDF_POINTER|FDF_R,  0, GET_ShadowRGB,    SET_ShadowRGB },
   { "Sunken",       0, FDF_LONG|FDF_RW,    0, GET_Sunken,       SET_Sunken },
   /*** Variable Fields ***/
   { "Height",       0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_Height,  SET_Height },
   { "Width",        0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_Width,   SET_Width },
   { "X",            0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_X,       SET_X },
   { "XCoord",       0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_X,       SET_X },
   { "XOffset",      0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_XOffset, SET_XOffset },
   { "Y",            0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_Y,       SET_Y },
   { "YCoord",       0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_Y,       SET_Y },
   { "YOffset",      0, FDF_VARIABLE|FDF_DOUBLE|FDF_PERCENTAGE|FDF_RW, 0, GET_YOffset, SET_YOffset },
   END_FIELD
};

static ERROR ARROW_Draw(struct Arrow *, APTR);
static ERROR ARROW_Free(struct Arrow *, APTR);
static ERROR ARROW_Hide(struct Arrow *, APTR);
static ERROR ARROW_Init(struct Arrow *, APTR);
static ERROR ARROW_NewObject(struct Arrow *, APTR);
static ERROR ARROW_Show(struct Arrow *, APTR);

static struct ActionArray ArrowActions[] = {
   { AC_Draw,         ARROW_Draw },
   { AC_Free,         ARROW_Free },
   { AC_Hide,         ARROW_Hide },
   { AC_Init,         ARROW_Init },
   { AC_NewObject,    ARROW_NewObject },
   { AC_Show,         ARROW_Show },
   { NULL, NULL }
};

/*****************************************************************************
** Command: Init()
*/

static ERROR CMDInit(OBJECTPTR argModule, struct KernelBase *argKernelBase)
{
   KernelBase = argKernelBase;

   if (CreateObject(ID_MODULE, NULL, &StringsMod, NULL, FID_Name|TSTRING, "strings", TAGEND) IS ERR_Okay) {
      if (GetField(StringsMod, FID_ModBase, FT_POINTER, &StringsBase) != ERR_Okay) {
         return(ObjectError(argModule, ERH_InitModule, ERR_GetField));
      }
   }
   else return(ObjectError(argModule, ERH_InitModule, ERR_CreateObject));

   return(CreateObject(ID_CLASS, NULL, (OBJECTPTR *)&ArrowClass, NULL,
      FID_BaseClassID|TLONG, ID_ARROW,
      FID_Version|TFLOAT,    VER_ARROW,
      FID_Name|TSTRING,      "Arrow",
      FID_Category|TLONG,    CCF_GUI,
      FID_Actions|TPTR,      ArrowActions,
      FID_Fields|TPTR,       ArrowFields,
      FID_Flags|TLONG,       CLF_PRIVATEONLY,
      FID_Size|TLONG,        sizeof(struct Arrow),
      TAGEND));
}

/*****************************************************************************
** Command: Expunge()
*/

static ERROR CMDExpunge(void)
{
   if (StringsMod) { Action(AC_Free, StringsMod, NULL); StringsMod = NULL; }
   if (ArrowClass) { Action(AC_Free, ArrowClass, NULL); ArrowClass = NULL; }
   return(ERR_Okay);
}

/*****************************************************************************
** Internal: DrawArrow()
*/

static void DrawArrow(struct Arrow *Self, struct Render *Drawable, struct Bitmap *Bitmap)
{
   struct Variable var;
   struct mtDrawRGBLine line;
   struct RGB *col1, *col2;
   LONG width, height, xcoord, ycoord, arrowx, arrowy, arrowwidth, arrowheight;
   WORD bmpopacity;

   if (!(Self->Flags & AF_VISIBLE)) return;

   if (Self->Frame) {
      if ((Drawable->Frame > 0) AND (Drawable->Frame != Self->Frame)) return;
   }

   var.Type = FD_DOUBLE;
   GET_X(Self, &var);      xcoord = var.Double;
   GET_Y(Self, &var);      ycoord = var.Double;
   GET_Width(Self, &var);  width  = var.Double * Self->Scale;
   GET_Height(Self, &var); height = var.Double * Self->Scale;

   if ((width <= 0) OR (height <= 0)) return;

   bmpopacity = Bitmap->Opacity;
   Bitmap->Opacity = Self->Opacity;

   arrowx      = Self->ArrowX;
   arrowy      = Self->ArrowY;
   arrowwidth  = Self->ArrowWidth;
   arrowheight = Self->ArrowHeight;

   /*** Determine the position of the image ***/

   if (Self->AlignFlags & ALIGN_HORIZONTAL) arrowx = xcoord + ((width - arrowwidth) / 2);
   else if (Self->AlignFlags & ALIGN_RIGHT) arrowx = xcoord + (width  - arrowwidth);
   else arrowx = Self->ArrowX + xcoord;

   if (Self->AlignFlags & ALIGN_VERTICAL) arrowy = ycoord + ((height - arrowheight) / 2);
   else if (Self->AlignFlags & ALIGN_BOTTOM) arrowy = ycoord + (height  - arrowheight);
   else arrowy = Self->ArrowY + ycoord;

   /*** Draw the arrow now ***/

   line.Red   = Self->Colour.Red;
   line.Green = Self->Colour.Green;
   line.Blue  = Self->Colour.Blue;

   if (Self->Direction IS ARROW_UP) {
      line.XCoord = arrowx + (arrowwidth>>1);
      line.YCoord = arrowy;
      line.XEnd   = arrowx;
      line.YEnd   = arrowy + arrowheight - 1;

      if (arrowwidth & 1) {
         while (line.XEnd < arrowx + arrowwidth) {
            Action(MT_DrawRGBLine, Bitmap, &line);
            line.XEnd++;
         }
      }
      else {
         line.XCoord--;
         while (line.XEnd <= line.XCoord) {
            Action(MT_DrawRGBLine, Bitmap, &line);
            line.XEnd++;
         }

         line.XCoord++;
         while (line.XEnd < arrowx + arrowwidth) {
            Action(MT_DrawRGBLine, Bitmap, &line);
            line.XEnd++;
         }
      }

      if (Self->Flags & (AF_RAISED|AF_SUNKEN)) {
         if (Self->Flags & AF_RAISED) {
            col1 = &Self->Highlight;
            col2 = &Self->Shadow;

            /*** Left line ***/
            line.Red    = col1->Red;
            line.Green  = col1->Green;
            line.Blue   = col1->Blue;
            line.XCoord = arrowx + (arrowwidth>>1);
            if (!(arrowwidth & 1)) line.XCoord--;
            line.YCoord = arrowy;
            line.XEnd   = arrowx;
            line.YEnd   = arrowy + arrowheight - 1;
            Action(MT_DrawRGBLine, Bitmap, &line);

            /*** Right line ***/
            line.Red    = col2->Red;
            line.Green  = col2->Green;
            line.Blue   = col2->Blue;
            line.XCoord = arrowx + (arrowwidth>>1);
            line.XEnd   = arrowx + arrowwidth - 1;
            line.YCoord = arrowy;
            line.YEnd   = arrowy + arrowheight - 1;
            Action(MT_DrawRGBLine, Bitmap, &line);
         }
         else {
            col1 = &Self->Shadow;
            col2 = &Self->Highlight;

            /*** Right line ***/
            line.Red    = col2->Red;
            line.Green  = col2->Green;
            line.Blue   = col2->Blue;
            line.XCoord = arrowx + (arrowwidth>>1);
            line.XEnd   = arrowx + arrowwidth - 1;
            line.YCoord = arrowy;
            line.YEnd   = arrowy + arrowheight - 1;
            Action(MT_DrawRGBLine, Bitmap, &line);

            /*** Left line ***/
            line.Red    = col1->Red;
            line.Green  = col1->Green;
            line.Blue   = col1->Blue;
            line.XCoord = arrowx + (arrowwidth>>1);
            if (!(arrowwidth & 1)) line.XCoord--;
            line.YCoord = arrowy;
            line.XEnd   = arrowx;
            line.YEnd   = arrowy + arrowheight - 1;
            Action(MT_DrawRGBLine, Bitmap, &line);
         }

         /*** Bottom line ***/
         line.Red    = col2->Red;
         line.Green  = col2->Green;
         line.Blue   = col2->Blue;
         line.XCoord = arrowx;
         line.YCoord = arrowy + arrowheight - 1;
         line.XEnd   = arrowx + arrowwidth - 1;
         line.YEnd   = arrowy + arrowheight - 1;
         Action(MT_DrawRGBLine, Bitmap, &line);
      }
   }
   else if (Self->Direction IS ARROW_DOWN) {
      line.XCoord = arrowx + (arrowwidth>>1);
      line.YCoord = arrowy + arrowheight - 1;
      line.XEnd   = arrowx;
      line.YEnd   = arrowy;

      if (arrowwidth & 1) {
         while (line.XEnd < arrowx + arrowwidth) {
            Action(MT_DrawRGBLine, Bitmap, &line);
            line.XEnd++;
         }
      }
      else {
         line.XCoord--;
         while (line.XEnd <= line.XCoord) {
            Action(MT_DrawRGBLine, Bitmap, &line);
            line.XEnd++;
         }

         line.XCoord++;
         while (line.XEnd < arrowx + arrowwidth) {
            Action(MT_DrawRGBLine, Bitmap, &line);
            line.XEnd++;
         }
      }

      if (Self->Flags & (AF_RAISED|AF_SUNKEN)) {
         if (Self->Flags & AF_RAISED) {
            col1 = &Self->Highlight;
            col2 = &Self->Shadow;

            /*** Left line ***/
            line.Red    = col1->Red;
            line.Green  = col1->Green;
            line.Blue   = col1->Blue;
            line.XCoord = arrowx + (arrowwidth>>1);
            if (!(arrowwidth & 1)) line.XCoord--;
            line.YCoord = arrowy + arrowheight - 1;
            line.XEnd   = arrowx;
            line.YEnd   = arrowy;
            Action(MT_DrawRGBLine, Bitmap, &line);

            /*** Right line ***/
            line.Red    = col2->Red;
            line.Green  = col2->Green;
            line.Blue   = col2->Blue;
            line.XCoord = arrowx + (arrowwidth>>1);
            line.XEnd   = arrowx + arrowwidth - 1;
            line.YCoord = arrowy + arrowheight - 1;
            line.YEnd   = arrowy;
            Action(MT_DrawRGBLine, Bitmap, &line);
         }
         else {
            col1 = &Self->Shadow;
            col2 = &Self->Highlight;

            /*** Right line ***/
            line.Red    = col2->Red;
            line.Green  = col2->Green;
            line.Blue   = col2->Blue;
            line.XCoord = arrowx + (arrowwidth>>1);
            line.XEnd   = arrowx + arrowwidth - 1;
            line.YCoord = arrowy + arrowheight - 1;
            line.YEnd   = arrowy;
            Action(MT_DrawRGBLine, Bitmap, &line);

            /*** Left line ***/
            line.Red    = col1->Red;
            line.Green  = col1->Green;
            line.Blue   = col1->Blue;
            line.XCoord = arrowx + (arrowwidth>>1);
            if (!(arrowwidth & 1)) line.XCoord--;
            line.XEnd   = arrowx;
            line.YCoord = arrowy + arrowheight - 1;
            line.YEnd   = arrowy;
            Action(MT_DrawRGBLine, Bitmap, &line);
         }

         /*** Top line ***/
         line.Red    = col2->Red;
         line.Green  = col2->Green;
         line.Blue   = col2->Blue;
         line.XCoord = arrowx;
         line.XEnd   = arrowx + arrowwidth - 1;
         line.YCoord = arrowy;
         Action(MT_DrawRGBLine, Bitmap, &line);
      }
   }
   else if (Self->Direction IS ARROW_LEFT) {
      line.XCoord = arrowx;
      line.YCoord = arrowy + (arrowheight>>1);
      line.XEnd   = arrowx + arrowwidth - 1;
      line.YEnd   = arrowy;

      if (arrowheight & 1) {
         while (line.YEnd < arrowy + arrowheight) {
            Action(MT_DrawRGBLine, Bitmap, &line);
            line.YEnd++;
         }
      }
      else {
         line.YCoord--;
         while (line.YEnd <= line.YCoord) {
            Action(MT_DrawRGBLine, Bitmap, &line);
            line.YEnd++;
         }

         line.YCoord++;
         while (line.YEnd < arrowy + arrowheight) {
            Action(MT_DrawRGBLine, Bitmap, &line);
            line.YEnd++;
         }
      }

      if (Self->Flags & (AF_RAISED|AF_SUNKEN)) {
         if (Self->Flags & AF_RAISED) {
            col1 = &Self->Highlight;
            col2 = &Self->Shadow;

            /*** Top line ***/
            line.Red    = col1->Red;
            line.Green  = col1->Green;
            line.Blue   = col1->Blue;
            line.XCoord = arrowx;
            line.YCoord = arrowy + (arrowheight>>1);
            if (!(arrowheight & 1)) line.YCoord--;
            line.XEnd   = arrowx + arrowwidth - 1;
            line.YEnd   = arrowy;
            Action(MT_DrawRGBLine, Bitmap, &line);

            /*** Bottom line ***/
            line.Red    = col2->Red;
            line.Green  = col2->Green;
            line.Blue   = col2->Blue;
            line.XCoord = arrowx;
            line.XEnd   = arrowx + arrowwidth - 1;
            line.YCoord = arrowy + (arrowheight>>1);
            line.YEnd   = arrowy + arrowheight - 1;
            Action(MT_DrawRGBLine, Bitmap, &line);
         }
         else {
            col1 = &Self->Shadow;
            col2 = &Self->Highlight;

            /*** Bottom line ***/
            line.Red    = col2->Red;
            line.Green  = col2->Green;
            line.Blue   = col2->Blue;
            line.XCoord = arrowx;
            line.XEnd   = arrowx + arrowwidth - 1;
            line.YCoord = arrowy + (arrowheight>>1);
            line.YEnd   = arrowy + arrowheight - 1;
            Action(MT_DrawRGBLine, Bitmap, &line);

            /*** Top line ***/
            line.Red    = col1->Red;
            line.Green  = col1->Green;
            line.Blue   = col1->Blue;
            line.XCoord = arrowx;
            line.YCoord = arrowy + (arrowheight>>1);
            if (!(arrowheight & 1)) line.YCoord--;
            line.XEnd   = arrowx + arrowwidth - 1;
            line.YEnd   = arrowy;
            Action(MT_DrawRGBLine, Bitmap, &line);
         }

         /*** Right line ***/
         line.Red    = col2->Red;
         line.Green  = col2->Green;
         line.Blue   = col2->Blue;
         line.XCoord = arrowx + arrowwidth - 1;
         line.XEnd   = arrowx + arrowwidth - 1;
         line.YCoord = arrowy;
         line.YEnd   = arrowy + arrowheight - 1;
         Action(MT_DrawRGBLine, Bitmap, &line);
      }
   }
   else if (Self->Direction IS ARROW_RIGHT) {
      line.XCoord = arrowx + arrowwidth - 1;
      line.YCoord = arrowy + (arrowheight>>1);
      line.XEnd   = arrowx;
      line.YEnd   = arrowy;

      if (arrowheight & 1) {
         while (line.YEnd < arrowy + arrowheight) {
            Action(MT_DrawRGBLine, Bitmap, &line);
            line.YEnd++;
         }
      }
      else {
         line.YCoord--;
         while (line.YEnd <= line.YCoord) {
            Action(MT_DrawRGBLine, Bitmap, &line);
            line.YEnd++;
         }

         line.YCoord++;
         while (line.YEnd < arrowy + arrowheight) {
            Action(MT_DrawRGBLine, Bitmap, &line);
            line.YEnd++;
         }
      }

      if (Self->Flags & (AF_RAISED|AF_SUNKEN)) {
         if (Self->Flags & AF_RAISED) {
            col1 = &Self->Highlight;
            col2 = &Self->Shadow;

            /*** Top line ***/
            line.Red    = col1->Red;
            line.Green  = col1->Green;
            line.Blue   = col1->Blue;
            line.XCoord = arrowx + arrowwidth - 1;
            line.YCoord = arrowy + (arrowheight>>1);
            if (!(arrowheight & 1)) line.YCoord--;
            line.XEnd   = arrowx;
            line.YEnd   = arrowy;
            Action(MT_DrawRGBLine, Bitmap, &line);

            /*** Bottom line ***/
            line.Red    = col2->Red;
            line.Green  = col2->Green;
            line.Blue   = col2->Blue;
            line.XCoord = arrowx + arrowwidth - 1;
            line.YCoord = arrowy + (arrowheight>>1);
            line.XEnd   = arrowx;
            line.YEnd   = arrowy + arrowheight - 1;
            Action(MT_DrawRGBLine, Bitmap, &line);

            line.Red    = col2->Red;
            line.Green  = col2->Green;
            line.Blue   = col2->Blue;
         }
         else {
            col1 = &Self->Shadow;
            col2 = &Self->Highlight;

            /*** Bottom line ***/
            line.Red    = col2->Red;
            line.Green  = col2->Green;
            line.Blue   = col2->Blue;
            line.XCoord = arrowx + arrowwidth - 1;
            line.YCoord = arrowy + (arrowheight>>1);
            line.XEnd   = arrowx;
            line.YEnd   = arrowy + arrowheight - 1;
            Action(MT_DrawRGBLine, Bitmap, &line);

            /*** Top line ***/
            line.Red    = col1->Red;
            line.Green  = col1->Green;
            line.Blue   = col1->Blue;
            line.XCoord = arrowx + arrowwidth - 1;
            line.YCoord = arrowy + (arrowheight>>1);
            if (!(arrowheight & 1)) line.YCoord--;
            line.XEnd   = arrowx;
            line.YEnd   = arrowy;
            Action(MT_DrawRGBLine, Bitmap, &line);

            line.Red    = col1->Red;
            line.Green  = col1->Green;
            line.Blue   = col1->Blue;
         }

         /*** Left line ***/
         line.XCoord = arrowx;
         line.XEnd   = arrowx;
         line.YCoord = arrowy;
         line.YEnd   = arrowy + arrowheight - 1;
         Action(MT_DrawRGBLine, Bitmap, &line);
      }
   }

   Bitmap->Opacity = bmpopacity;
}

/*****************************************************************************
** Self: Draw
*/

static ERROR ARROW_Draw(struct Arrow *Self, APTR Void)
{
   return(ActionMsg(AC_Draw, Self->DrawableID, NULL));
}

/*****************************************************************************
** Self: Free
*/

static ERROR ARROW_Free(struct Arrow *Self, APTR Void)
{
   /*** Remove all subscriptions to our drawable container ***/

   if (Self->DrawableID) {
      OBJECTPTR drawable;
      if (AccessObject(Self->DrawableID, 5000, &drawable) IS ERR_Okay) {
         ActionTags(MT_SetRenderCallback, drawable, Self, NULL);
         ReleaseObject(drawable);
      }
   }

   return(ERR_Okay);
}

/*****************************************************************************
** Arrow: Hide
*/

static ERROR ARROW_Hide(struct Arrow *Self, APTR Void)
{
   if (Self->Flags & AF_VISIBLE) {
      Self->Flags &= ~AF_VISIBLE;
      if (Self->DrawableID) ActionMsg(AC_Draw, Self->DrawableID, NULL);
   }
   return(ERR_Okay);
}

/*****************************************************************************
** Arrow: Init
*/

static ERROR ARROW_Init(struct Arrow *Self, APTR Void)
{
   struct Variable value;
   OBJECTPTR drawable;

   /*** Find the drawable object that we are associated with ***/

   if (!Self->DrawableID) {
      OBJECTID OwnerID;
      OwnerID = GetOwner(Self);
      while ((OwnerID) AND (GetClassID(OwnerID) != ID_RENDER)) {
         OwnerID = GetOwnerID(OwnerID);
      }
      if (!OwnerID) return(ObjectError(Self, ERH_Init, ERR_UnsupportedOwner));
      else Self->DrawableID = OwnerID;
   }

   if (AccessObject(Self->DrawableID, 5000, &drawable) IS ERR_Okay) {
      ActionTags(MT_SetRenderCallback, drawable, Self, &DrawArrow);
      ReleaseObject(drawable);
   }
   else return(ObjectError(Self, ERH_Init, ERR_ExclusiveDenied));

   /* If dimension settings are missing (e.g. if it is impossible to
   ** determine width, height or a coordinate), then we set the missing
   ** fields to maximum possible values.
   */

   value.Type = FD_DOUBLE;
   value.Double = 0;

   if (!(Self->Dimensions & (DMF_FIXED_X|DMF_RELATIVE_X|DMF_FIXED_XOFFSET|DMF_RELATIVE_XOFFSET))) {
      SET_X(Self, &value);
   }

   if (!(Self->Dimensions & (DMF_FIXED_Y|DMF_RELATIVE_Y|DMF_FIXED_YOFFSET|DMF_RELATIVE_YOFFSET))) {
      SET_Y(Self, &value);
   }

   if (!(Self->Dimensions & (DMF_FIXED_WIDTH|DMF_RELATIVE_WIDTH|DMF_FIXED_XOFFSET|DMF_RELATIVE_XOFFSET))) {
      SET_XOffset(Self, &value);
   }

   if (!(Self->Dimensions & (DMF_FIXED_HEIGHT|DMF_RELATIVE_HEIGHT|DMF_FIXED_YOFFSET|DMF_RELATIVE_YOFFSET))) {
      SET_YOffset(Self, &value);
   }

   if (!Self->ArrowWidth) {
      value.Type = FD_LARGE;
      GET_Width(Self, &value); Self->ArrowWidth = value.Large;
   }

   if (!Self->ArrowHeight) {
      value.Type = FD_LARGE;
      GET_Height(Self, &value); Self->ArrowHeight = value.Large;
   }

   return(ERR_Okay);
}

/*****************************************************************************
** Arrow: NewObject
*/

static ERROR ARROW_NewObject(struct Arrow *Self, APTR Void)
{
   /*** Shadow colour ***/
   Self->Shadow.Red   = 64;
   Self->Shadow.Green = 64;
   Self->Shadow.Blue  = 64;

   /*** Internal colour ***/
   Self->Colour.Red   = 0;
   Self->Colour.Green = 0;
   Self->Colour.Blue  = 0;

   /*** Highlight colour ***/
   Self->Highlight.Red   = 192;
   Self->Highlight.Green = 192;
   Self->Highlight.Blue  = 192;

   Self->Opacity   = 255;
   Self->Scale     = 1;
   Self->Direction = ARROW_UP;
   Self->Flags    |= AF_VISIBLE;
   return(ERR_Okay);
}

/*****************************************************************************
** Arrow: Show
*/

static ERROR ARROW_Show(struct Arrow *Self, APTR Void)
{
   if (!(Self->Flags & AF_VISIBLE)) {
      Self->Flags |= AF_VISIBLE;
      if (Self->DrawableID) ActionMsg(AC_Draw, Self->DrawableID, NULL);
   }
   return(ERR_Okay);
}

/*****************************************************************************
** Function: ReadByte()
** Short:    Reads any double digit hex value and converts it into a number.
*/

static WORD ReadByte(STRING str)
{
   WORD Value;
   if (!*str) return(0);
   if ((*str >= '0') AND (*str <= '9')) Value = (*str - '0') * 16;
   else if ((*str >= 'A') AND (*str <= 'F')) Value = ((*str - 'A')+10) * 16;
   else if ((*str >= 'a') AND (*str <= 'f')) Value = ((*str - 'a')+10) * 16;
   else Value = 0;
   str++;
   if ((*str >= '0') AND (*str <= '9')) Value += (*str - '0');
   else if ((*str >= 'A') AND (*str <= 'F')) Value += (*str - 'A') + 10;
   else if ((*str >= 'a') AND (*str <= 'f')) Value += (*str - 'a') + 10;
   return(Value);
}

/*****************************************************************************

###FIELD###
Name:   Align
Short:  This field allows you to align an arrow within its drawable area.
Type:   LONG/FLAGS
Prefix: ALIGN
Status: Read/Write

###DESCRIPTION###
<p>If you would like to set an abstract position for an arrow, you can
give it an alignment.  Note that setting the alignment over-rides any
settings in related coordinate fields.  Valid alignment flags are as
follows:</p>

<blockquote>
<p>BOTTOM<br>
CENTER/MIDDLE<br>
LEFT<br>
HORIZONTAL<br>
RIGHT<br>
TOP<br>
VERTICAL</p>
</blockquote>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Colour
Short:  String-based field for setting the internal colour of an Arrow.
Type:   STRING
Status: Write

###DESCRIPTION###
<p>The colour of an Arrow can be declared by writing to this field.  The
colour must be in hexadecimal or separated-decimal format - for example
to create a pure red colour, a setting of "#ff0000" or "255,0,0" would be
valid.</p>

###SEE ALSO###
Field: ColourRGB

###END###

*****************************************************************************/

static ERROR SET_Colour(struct Arrow *Self, STRING Colour)
{
   if (*Colour IS '#') {
      Colour++;
      Self->Colour.Red   = ReadByte(Colour);
      if (*Colour) Colour++; if (*Colour) Colour++;
      Self->Colour.Green = ReadByte(Colour);
      if (*Colour) Colour++; if (*Colour) Colour++;
      Self->Colour.Blue  = ReadByte(Colour);
   }
   else {
      Self->Colour.Red = StrToInt(Colour);
      while ((*Colour) AND (*Colour != ',')) Colour++; if (*Colour) Colour++;
      Self->Colour.Green = StrToInt(Colour);
      while ((*Colour) AND (*Colour != ',')) Colour++; if (*Colour) Colour++;
      Self->Colour.Blue = StrToInt(Colour);
   }

   Self->Fill = TRUE;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   ColourRGB
Short:  Defines the internal colour of the arrow in RGB format.
Type:   *RGB
Status: Get/Set

###DESCRIPTION###
<p>The colour of an arrow can be read or written in RGB format using this
field.</p>

###SEE ALSO###
Field: Colour

###END###

*****************************************************************************/

static ERROR GET_ColourRGB(struct Arrow *Self, struct RGB *RGB)
{
   RGB->Red   = Self->Colour.Red;
   RGB->Green = Self->Colour.Green;
   RGB->Blue  = Self->Colour.Blue;
   return(ERR_Okay);
}

static ERROR SET_ColourRGB(struct Arrow *Self, struct RGB *RGB)
{
   if (RGB) {
      Self->Colour.Blue  = RGB->Red;
      Self->Colour.Green = RGB->Green;
      Self->Colour.Red   = RGB->Blue;
      return(ERR_Okay);
   }
   else return(ERR_Args);
}

/*****************************************************************************

###FIELD###
Name:   Direction
Short:  Determines the direction of the Arrow.
Type:   INDEX
Status: Read/Write

###DESCRIPTION###
<p>You can set the direction of an Arrow by writing to this field.  The
following index table illustrates the valid direction values:</p>

<pre>
  <b>Name</b>     <b>Index</b>         <b>Value</b>
  UP       ARROW_UP        1
  DOWN     ARROW_DOWN      2
  LEFT     ARROW_LEFT      3
  RIGHT    ARROW_RIGHT     4
</pre>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Frame
Short:  Determines the rendering frame that the Arrow will be applied to.
Type:   LONG
Status: Read/Write

###DESCRIPTION###
<p>Setting the Frame field to any value other than zero will force the Arrow
to be drawn only when the Render frame matches the specified value.  For
instance, if the Render container has a Frame setting of 2, and the Arrow has
a Frame of 1, then no graphics will be drawn to represent the Arrow as the
numbers do not match.</p>

###END###

*****************************************************************************/

/*****************************************************************************

###FIELD###
Name:   Highlight
Short:  String-based field for setting the highlight colour of an arrow object.
Type:   STRING
Status: Write

###DESCRIPTION###
<p>If an Arrow object should be highlighted around the edges, set this field
to the desired colour for the highlighting.  The colour specification must be
in the standard hexadecimal or decimal separated format.</p>

<p>Passing a NULL value to this field will turn off highlighting.</p>

###SEE ALSO###
Field: HighlightRGB, Shadow

###END###

*****************************************************************************/

static ERROR SET_Highlight(struct Arrow *Self, STRING Colour)
{
   if (!Colour) {
      Self->Thickness = 0;
      return(ERR_Okay);
   }

   if (*Colour IS '#') {
      Colour++;
      Self->Highlight.Red   = ReadByte(Colour);
      if (*Colour) Colour++; if (*Colour) Colour++;
      Self->Highlight.Green = ReadByte(Colour);
      if (*Colour) Colour++; if (*Colour) Colour++;
      Self->Highlight.Blue  = ReadByte(Colour);
   }
   else {
      Self->Highlight.Red = StrToInt(Colour);
      while ((*Colour) AND (*Colour != ',')) Colour++; if (*Colour) Colour++;
      Self->Highlight.Green = StrToInt(Colour);
      while ((*Colour) AND (*Colour != ',')) Colour++; if (*Colour) Colour++;
      Self->Highlight.Blue = StrToInt(Colour);
   }

   if (!Self->Thickness) Self->Thickness = 1;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   HighlightRGB
Short:  Defines the highlight colour of the arrow in RGB format.
Type:   *RGB
Status: Read/Write

###DESCRIPTION###
<p>The highlight colour of an Arrow can be manipulated in RGB format using
this field.</p>

###SEE ALSO###
Fields: Highlight, ShadowRGB

###END###

*****************************************************************************/

static ERROR GET_HighlightRGB(struct Arrow *Self, struct RGB *Colour)
{
   Colour->Red   = Self->Highlight.Red;
   Colour->Green = Self->Highlight.Green;
   Colour->Blue  = Self->Highlight.Blue;
   return(ERR_Okay);
}

static ERROR SET_HighlightRGB(struct Arrow *Self, struct RGB *RGB)
{
   Self->Highlight.Blue  = RGB->Red;
   Self->Highlight.Green = RGB->Green;
   Self->Highlight.Red   = RGB->Blue;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Raised
Short:  If set to TRUE the arrow will appear to be raised into the foreground.
Type:   BOOLEAN
Status: Read/Write

###DESCRIPTION###
<p>If you have set the Highlight and Shadow fields of an Arrow object then
you will need to decide whether or not the Arrow should be given a sunken
or raised effect when it is drawn.  To give it a raised effect you will need
to set this field to TRUE, if not then you should set the Sunken field.</p>

###SEE ALSO###
Field: Sunken

###END###

*****************************************************************************/

static ERROR GET_Raised(struct Arrow *Self, LONG *Value)
{
   if (Self->Flags & AF_RAISED) *Value = TRUE;
   else *Value = FALSE;
   return(ERR_Okay);
}

static ERROR SET_Raised(struct Arrow *Self, LONG *Value)
{
   if (*Value) {
      if (Self->Flags & AF_RAISED) return(ERR_Okay);
      Self->Flags = (Self->Flags & ~AF_SUNKEN) | AF_RAISED;
   }
   else Self->Flags &= ~AF_RAISED;
   if (Self->Head.ObjectFlags & NF_INITIALISED) ActionMsg(AC_Draw, Self->DrawableID, NULL);
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Scale
Short:  This field alters the width and height dimensions through scaling.
Type:   FLOAT
Status: Read/Write

###DESCRIPTION###
<p>The scale of an Arrow can be changed by writing a value to this
field.  Both the width and height will be adjusted by the factor that you
set, using the calculation "ScaledValue = Value * Scale".  The default
scale of 1 will leave the width and height unchanged, 0.5 would halve the
size of the arrow and a scale of 2 would double its size.</p>

<p>The Scale is limited to a range between 0 and 1000.</p>

###END###

*****************************************************************************/

static ERROR SET_Scale(struct Arrow *Self, FLOAT *Scale)
{
   if ((*Scale > 0) AND (*Scale < 1000)) {
      Self->Scale = *Scale;
      if (Self->DrawableID) ActionMsg(AC_Draw, Self->DrawableID, NULL);
      return(ERR_Okay);
   }
   else return(ERR_OutOfRange);
}

/*****************************************************************************

###FIELD###
Name:   Shadow
Short:  String-based field for setting the shadow colour of an Arrow object.
Type:   STRING
Status: Write

###DESCRIPTION###
<p>If an Arrow object should be shadowed around the edges, set this field
to the desired colour for the shadowing.  The colour specification must be
in the standard hexadecimal or decimal separated format.</p>

<p>Passing a NULL value to this field will turn off shadowing.</p>

###SEE ALSO###
Field: ShadowRGB, Highlight

###END###

*****************************************************************************/

static ERROR SET_Shadow(struct Arrow *Self, STRING Colour)
{
   if (*Colour IS '#') {
      Colour++;
      Self->Shadow.Red   = ReadByte(Colour);
      if (*Colour) Colour++; if (*Colour) Colour++;
      Self->Shadow.Green = ReadByte(Colour);
      if (*Colour) Colour++; if (*Colour) Colour++;
      Self->Shadow.Blue  = ReadByte(Colour);
   }
   else {
      Self->Shadow.Red = StrToInt(Colour);
      while ((*Colour) AND (*Colour != ',')) Colour++; if (*Colour) Colour++;
      Self->Shadow.Green = StrToInt(Colour);
      while ((*Colour) AND (*Colour != ',')) Colour++; if (*Colour) Colour++;
      Self->Shadow.Blue = StrToInt(Colour);
   }

   if (!Self->Thickness) Self->Thickness = 1;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Height
Short:  The height of the Arrow.
Type:   FLOAT/PERCENTAGE
Status: Read/Write

###DESCRIPTION###
<p>To set the height of an Arrow, write a value to this field.  A percentage
may be specified if the FD_PERCENT flag is used, otherwise the setting is
taken as being fixed.  As an alternative to setting the Height, you can
otherwise set the YCoord field in conjunction with the YOffset to create a
dynamically calculated height value.</p>

###SEE ALSO###
Field: Width, YCoord, YOffset

###END###

*****************************************************************************/

static ERROR GET_Height(struct Arrow *Self, struct Variable *Value)
{
   OBJECTPTR drawable;
   DOUBLE containerheight, ycoord, value;

   if (AccessObject(Self->DrawableID, 1000, &drawable) IS ERR_Okay) {
      GetField(drawable, FID_Height, FT_DOUBLE, &containerheight);
      ReleaseObject(drawable);

      if (containerheight <= 0) return(ObjectError(Self, ERH_GetField, ERR_GetField));

      if (Self->Dimensions & DMF_FIXED_HEIGHT) value = Self->Height;
      else if (Self->Dimensions & DMF_RELATIVE_HEIGHT) {
         value = Self->Height * containerheight / 100;
      }
      else if ((Self->Dimensions & (DMF_FIXED_Y|DMF_RELATIVE_Y)) AND
               (Self->Dimensions & (DMF_FIXED_YOFFSET|DMF_RELATIVE_YOFFSET))) {
         if (Self->Dimensions & DMF_FIXED_Y) ycoord = Self->YCoord;
         else ycoord = containerheight * Self->YCoord / 100;
         if (Self->Dimensions & DMF_FIXED_YOFFSET) value = containerheight - ycoord - Self->YOffset;
         else value = containerheight - ycoord - (containerheight * Self->YOffset / 100);
      }
      else value = 0;

      if (Value->Type & FD_PERCENTAGE) value = (value * 100) / containerheight;
   }
   else return(ObjectError(Self, ERH_GetField, ERR_AccessObject));

   if (Value->Type & FD_DOUBLE) Value->Double = value;
   else if (Value->Type & FD_LARGE) Value->Large = value;
   else return(ObjectError(Self, ERH_GetField, ERR_FieldTypeMismatch));
   return(ERR_Okay);
}

static ERROR SET_Height(struct Arrow *Self, struct Variable *Value)
{
   if (Value->Type & FD_DOUBLE) {
      if (Value->Double <= 0) {
         DPrintF("@SetField:","[Arrow] A height of %.2f is illegal.", Value->Double);
         return(ERR_OutOfRange);
      }
      Self->Height = Value->Double;
   }
   else if (Value->Type & FD_LARGE) {
      if (Value->Large <= 0) {
         DPrintF("@SetField:","[Arrow] A height of %d is illegal.", Value->Large);
         return(ERR_OutOfRange);
      }
      Self->Height = Value->Large;
   }
   else return(ObjectError(Self, ERH_GetField, ERR_FieldTypeMismatch));

   if (Value->Type & FD_PERCENTAGE) Self->Dimensions = (Self->Dimensions & ~DMF_FIXED_HEIGHT) | DMF_RELATIVE_HEIGHT;
   else Self->Dimensions = (Self->Dimensions & ~DMF_RELATIVE_HEIGHT) | DMF_FIXED_HEIGHT;
   return(ERR_Okay);
}
/****************************************************************************

###FIELD###
Name:   Opacity
Short:  Determines the level of translucency applied to an arrow.
Type:   FLOAT
Status: Get/Set

###DESCRIPTION###
<p>This field determines the translucency level of an arrow graphic.  The
default setting is 100%, which means that the arrow will be solid.  Any
other value that you set here will alter the impact of am arrow graphic over
the destination drawable.  High values will retain the boldness of the arrow,
while low values can render it close to invisible.</p>

<p>Please note that the use of translucency will always have an impact on the
time it normally takes to draw an arrow.</p>

###END###

****************************************************************************/

static ERROR GET_Opacity(struct Arrow *Self, FLOAT *Value)
{
   *Value = Self->Opacity * 100 / 255;
   return(ERR_Okay);
}

static ERROR SET_Opacity(struct Arrow *Self, FLOAT *Value)
{
   if (*Value < 0) Self->Opacity = 0;
   else if (*Value > 100) Self->Opacity = 255;
   else Self->Opacity = *Value * 255 / 100;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   ShadowRGB
Short:  Defines the shadow colour of the Arrow in RGB format.
Type:   *RGB
Status: Read/Write

###DESCRIPTION###
<p>The shadow colour of an Arrow can be manipulated in RGB format using
this field.</p>

###SEE ALSO###
Fields: HighlightRGB, Shadow

###END###

*****************************************************************************/

static ERROR GET_ShadowRGB(struct Arrow *Self, struct RGB *Colour)
{
   Colour->Red   = Self->Shadow.Red;
   Colour->Green = Self->Shadow.Green;
   Colour->Blue  = Self->Shadow.Blue;
   return(ERR_Okay);
}

static ERROR SET_ShadowRGB(struct Arrow *Self, struct RGB *RGB)
{
   Self->Shadow.Blue  = RGB->Red;
   Self->Shadow.Green = RGB->Green;
   Self->Shadow.Red   = RGB->Blue;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Sunken
Short:  Set to TRUE to make the arrow appear to sink into the background.
Type:   BOOLEAN
Status: Read/Write

###DESCRIPTION###
<p>If you have set the Highlight and Shadow fields of an Arrow object then you
will need to decide whether or not the Arrow should be given a sunken or raised
effect when it is drawn.  To give it a sunken effect you will need to set this
field to TRUE, if not then you should set the Raised field.</p>

###SEE ALSO###
Field: Raised

###END###

*****************************************************************************/

static ERROR GET_Sunken(struct Arrow *Self, LONG *Value)
{
   if (Self->Flags & AF_SUNKEN) *Value = TRUE;
   else *Value = FALSE;
   return(ERR_Okay);
}

static ERROR SET_Sunken(struct Arrow *Self, LONG *Value)
{
   if (*Value) {
      if (Self->Flags & AF_SUNKEN) return(ERR_Okay);
      Self->Flags = (Self->Flags & ~AF_RAISED) | AF_SUNKEN;
   }
   else Self->Flags &= ~AF_SUNKEN;
   if (Self->Head.ObjectFlags & NF_INITIALISED) ActionMsg(AC_Draw, Self->DrawableID, NULL);
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   Width
Short:  The width of the Arrow.
Type:   FLOAT/PERCENTAGE
Status: Read/Write

###DESCRIPTION###
<p>To set the width of an Arrow, write a value to this field.  A percentage
may be specified if the FD_PERCENT flag is used, otherwise the setting is
taken as being fixed.  As an alternative to setting the Width, you can
otherwise set the XCoord field in conjunction with the XOffset to create a
dynamically calculated width value.</p>

###SEE ALSO###
Field: Height, XCoord, XOffset

###END###

*****************************************************************************/

static ERROR GET_Width(struct Arrow *Self, struct Variable *Value)
{
   OBJECTPTR drawable;
   DOUBLE containerwidth, xcoord, value;

   if (AccessObject(Self->DrawableID, 1000, &drawable) IS ERR_Okay) {
      GetField(drawable, FID_Width, FT_DOUBLE, &containerwidth);
      ReleaseObject(drawable);

      if (containerwidth <= 0) return(ObjectError(Self, ERH_GetField, ERR_GetField));

      if (Self->Dimensions & DMF_FIXED_WIDTH) value = Self->Width;
      else if (Self->Dimensions & DMF_RELATIVE_WIDTH) {
         value = Self->Width * containerwidth / 100;
      }
      else if ((Self->Dimensions & (DMF_FIXED_X|DMF_RELATIVE_X)) AND
               (Self->Dimensions & (DMF_FIXED_XOFFSET|DMF_RELATIVE_XOFFSET))) {
         if (Self->Dimensions & DMF_FIXED_X) xcoord = Self->XCoord;
         else xcoord = containerwidth * Self->XCoord / 100;
         if (Self->Dimensions & DMF_FIXED_XOFFSET) value = containerwidth - xcoord - Self->XOffset;
         else value = containerwidth - xcoord - (containerwidth * Self->XOffset / 100);
      }
      else value = 0;

      if (Value->Type & FD_PERCENTAGE) value = (value * 100) / containerwidth;
   }
   else return(ObjectError(Self, ERH_GetField, ERR_AccessObject));

   if (Value->Type & FD_DOUBLE) Value->Double = value;
   else if (Value->Type & FD_LARGE) Value->Large = value;
   else return(ObjectError(Self, ERH_GetField, ERR_FieldTypeMismatch));
   return(ERR_Okay);
}

static ERROR SET_Width(struct Arrow *Self, struct Variable *Value)
{
   if (Value->Type & FD_DOUBLE) {
      if (Value->Double <= 0) {
         DPrintF("@SetField:","[Arrow] A width of %.2f is illegal.", Value->Double);
         return(ERR_OutOfRange);
      }
      Self->Width = Value->Double;
   }
   else if (Value->Type & FD_LARGE) {
      if (Value->Large <= 0) {
         DPrintF("@SetField:","[Arrow] A width of %d is illegal.", Value->Large);
         return(ERR_OutOfRange);
      }
      Self->Width = Value->Large;
   }
   else return(ObjectError(Self, ERH_SetField, ERR_FieldTypeMismatch));

   if (Value->Type & FD_PERCENTAGE) Self->Dimensions = (Self->Dimensions & ~DMF_FIXED_WIDTH) | DMF_RELATIVE_WIDTH;
   else Self->Dimensions = (Self->Dimensions & ~DMF_RELATIVE_WIDTH) | DMF_FIXED_WIDTH;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   XCoord
Short:  The horizontal position of the Arrow.
Type:   FLOAT/PERCENTAGE
Status: Read/Write

###DESCRIPTION###
<p>The horizontal position of an Arrow can be fixed to an absolute coordinate by
writing a value to the XCoord field.  If you use the FD_PERCENT flag then
you may write a relative/percentage based value to the field.  Negative
values are permitted.</p>

###SEE ALSO###
Field: XOffset, YCoord

###END###

*****************************************************************************/

static ERROR GET_X(struct Arrow *Self, struct Variable *Value)
{
   OBJECTPTR drawable;
   DOUBLE containerwidth, width, value;

   if (AccessObject(Self->DrawableID, 1000, &drawable) IS ERR_Okay) {
      GetField(drawable, FID_Width, FT_DOUBLE, &containerwidth);
      ReleaseObject(drawable);

      if (containerwidth <= 0) return(ERR_GetField);

      if (Self->Dimensions & DMF_FIXED_X) value = Self->XCoord;
      else if (Self->Dimensions & DMF_RELATIVE_X) {
         value = Self->XCoord * containerwidth / 100;
      }
      else if ((Self->Dimensions & (DMF_FIXED_WIDTH|DMF_RELATIVE_WIDTH)) AND
               (Self->Dimensions & (DMF_FIXED_XOFFSET|DMF_RELATIVE_XOFFSET))) {
         if (Self->Dimensions & DMF_FIXED_WIDTH) width = Self->Width;
         else width = containerwidth * Self->Width / 100;
         if (Self->Dimensions & DMF_FIXED_XOFFSET) value = containerwidth - width - Self->XOffset;
         else value = containerwidth - width - (containerwidth * Self->XOffset / 100);
      }
      else value = 0;

      if (Value->Type & FD_PERCENTAGE) value = (value * 100) / containerwidth;
   }
   else return(ObjectError(Self, ERH_GetField, ERR_ExclusiveDenied));

   if (Value->Type & FD_DOUBLE) Value->Double = value;
   else if (Value->Type & FD_LARGE) Value->Large = value;
   else return(ObjectError(Self, ERH_GetField, ERR_FieldTypeMismatch));
   return(ERR_Okay);
}

static ERROR SET_X(struct Arrow *Self, struct Variable *Value)
{
   if (Value->Type & FD_DOUBLE) Self->XCoord = Value->Double;
   else if (Value->Type & FD_LARGE) Self->XCoord = Value->Large;
   else return(ObjectError(Self, ERH_SetField, ERR_FieldTypeMismatch));

   if (Value->Type & FD_PERCENTAGE) Self->Dimensions = (Self->Dimensions & ~DMF_FIXED_X) | DMF_RELATIVE_X;
   else Self->Dimensions = (Self->Dimensions & ~DMF_RELATIVE_X) | DMF_FIXED_X;
   return(ERR_Okay);
}


/*****************************************************************************

###FIELD###
Name:   XOffset
Short:  The horziontal offset for the Arrow.
Type:   FLOAT/PERCENTAGE
Status: Read/Write

###DESCRIPTION###
<p>The XOffset has a dual purpose depending on whether or not it is set in
conjunction with an X coordinate or a Width based field.</p>

<p>If set in conjunction with an X coordinate then the Arrow will be drawn
from that X coordinate up to the width of the container, minus the value
given in the XOffset.  This means that the width of the Arrow is
dynamically calculated in relation to the width of the container.</p>

<p>If the XOffset field is set in conjunction with a fixed or relative width
then the Arrow will be drawn at an X coordinate calculated from the formula
"XCoord = ContainerWidth - ArrowWidth - XOffset".</p>

###SEE ALSO###
Field: XCoord, Width, YOffset

###END###

*****************************************************************************/

static ERROR GET_XOffset(struct Arrow *Self, struct Variable *Value)
{
   OBJECTPTR drawable;
   DOUBLE containerwidth, width, value;

   if (AccessObject(Self->DrawableID, 1000, &drawable) IS ERR_Okay) {
      GetField(drawable, FID_Width, FT_DOUBLE, &containerwidth);
      ReleaseObject(drawable);

      if (containerwidth <= 0) return(ERR_GetField);

      if (Self->Dimensions & DMF_FIXED_XOFFSET) value = Self->XOffset;
      else if (Self->Dimensions & DMF_RELATIVE_XOFFSET) {
         value = Self->XOffset * containerwidth / 100;
      }
      else if ((Self->Dimensions & (DMF_FIXED_X|DMF_RELATIVE_X)) AND
               (Self->Dimensions & (DMF_FIXED_WIDTH|DMF_RELATIVE_WIDTH))) {
         if (Self->Dimensions & DMF_FIXED_WIDTH) width = Self->Width;
         else width = containerwidth * Self->Width / 100;
         if (Self->Dimensions & DMF_FIXED_X) value = containerwidth - (Self->XCoord + width);
         else value = containerwidth - ((Self->XCoord * containerwidth / 100) + width);
      }
      else value = 0;

      if (Value->Type & FD_PERCENTAGE) value = (value * 100) / containerwidth;
   }
   else return(ObjectError(Self, ERH_GetField, ERR_ExclusiveDenied));

   if (Value->Type & FD_DOUBLE) Value->Double = value;
   else if (Value->Type & FD_LARGE) Value->Large = value;
   else return(ObjectError(Self, ERH_GetField, ERR_FieldTypeMismatch));
   return(ERR_Okay);
}

static ERROR SET_XOffset(struct Arrow *Self, struct Variable *Value)
{
   if (Value->Type & FD_DOUBLE) Self->XOffset= Value->Double;
   else if (Value->Type & FD_LARGE) Self->XOffset= Value->Large;
   else return(ObjectError(Self, ERH_SetField, ERR_FieldTypeMismatch));

   if (Value->Type & FD_PERCENTAGE) Self->Dimensions = (Self->Dimensions & ~DMF_FIXED_XOFFSET) | DMF_RELATIVE_XOFFSET;
   else Self->Dimensions = (Self->Dimensions & ~DMF_RELATIVE_XOFFSET) | DMF_FIXED_XOFFSET;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   YCoord
Short:  The vertical position of the Arrow..
Type:   FLOAT/PERCENTAGE
Status: Read/Write

###DESCRIPTION###
<p>The vertical position of an Arrow can be fixed to an absolute coordinate by
writing a value to the YCoord field.  If you use the FD_PERCENT flag then
you may write a relative/percentage based value to the field.  Negative
values are permitted.</p>

###SEE ALSO###
Field: YOffset, XCoord

###END###

*****************************************************************************/

static ERROR GET_Y(struct Arrow *Self, struct Variable *Value)
{
   OBJECTPTR drawable;
   DOUBLE value, containerheight, height;

   if (AccessObject(Self->DrawableID, 1000, &drawable) IS ERR_Okay) {
      GetField(drawable, FID_Height, FT_DOUBLE, &containerheight);
      ReleaseObject(drawable);

      if (containerheight <= 0) return(ERR_GetField);

      if (Self->Dimensions & DMF_FIXED_Y) value = Self->YCoord;
      else if (Self->Dimensions & DMF_RELATIVE_Y) {
         value = Self->YCoord * containerheight / 100;
      }
      else if ((Self->Dimensions & (DMF_FIXED_HEIGHT|DMF_RELATIVE_HEIGHT)) AND
               (Self->Dimensions & (DMF_FIXED_YOFFSET|DMF_RELATIVE_YOFFSET))) {
         if (Self->Dimensions & DMF_FIXED_HEIGHT) height = Self->Height;
         else height = containerheight * Self->Height / 100;
         if (Self->Dimensions & DMF_FIXED_YOFFSET) value = containerheight - height - Self->YOffset;
         else value = containerheight - height - (containerheight * Self->YOffset / 100);
      }
      else value = 0;

      if (Value->Type & FD_PERCENTAGE) value = (value * 100) / containerheight;
   }
   else return(ObjectError(Self, ERH_GetField, ERR_AccessObject));

   if (Value->Type & FD_DOUBLE) Value->Double = value;
   else if (Value->Type & FD_LARGE) Value->Large = value;
   else return(ObjectError(Self, ERH_GetField, ERR_FieldTypeMismatch));
   return(ERR_Okay);
}

static ERROR SET_Y(struct Arrow *Self, struct Variable *Value)
{
   if (Value->Type & FD_DOUBLE) Self->YCoord = Value->Double;
   else if (Value->Type & FD_LARGE) Self->YCoord = Value->Large;
   else return(ObjectError(Self, ERH_SetField, ERR_FieldTypeMismatch));

   if (Value->Type & FD_PERCENTAGE) Self->Dimensions = (Self->Dimensions & ~DMF_FIXED_Y) | DMF_RELATIVE_Y;
   else Self->Dimensions = (Self->Dimensions & ~DMF_RELATIVE_Y) | DMF_FIXED_Y;
   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   YOffset
Short:  The vertical offset for the Arrow.
Type:   FLOAT/PERCENTAGE
Status: Read/Write

###DESCRIPTION###
<p>The YOffset has a dual purpose depending on whether or not it is set in
conjunction with a Y coordinate or a Height based field.</p>

<p>If set in conjunction with a Y coordinate then the Arrow will be drawn
from that Y coordinate up to the height of the container, minus the value
given in the YOffset.  This means that the height of the Arrow is
dynamically calculated in relation to the height of the container.</p>

<p>If the YOffset field is set in conjunction with a fixed or relative height
then the Arrow will be drawn at a Y coordinate calculated from the formula
"YCoord = ContainerHeight - ArrowHeight - YOffset".</p>

###SEE ALSO###
Field: YCoord, Height, XOffset

###END###

*****************************************************************************/

static ERROR GET_YOffset(struct Arrow *Self, struct Variable *Value)
{
   OBJECTPTR drawable;
   DOUBLE value, containerheight, height;

   value = 0;
   if (AccessObject(Self->DrawableID, 1000, &drawable) IS ERR_Okay) {
      GetField(drawable, FID_Height, FT_DOUBLE, &containerheight);
      ReleaseObject(drawable);

      if (containerheight <= 0) return(ERR_GetField);

      if (Self->Dimensions & DMF_FIXED_YOFFSET) value = Self->YOffset;
      else if (Self->Dimensions & DMF_RELATIVE_YOFFSET) {
         value = Self->YOffset * containerheight / 100;
      }
      else if ((Self->Dimensions & (DMF_FIXED_Y|DMF_RELATIVE_Y)) AND
               (Self->Dimensions & (DMF_FIXED_HEIGHT|DMF_RELATIVE_HEIGHT))) {
         if (Self->Dimensions & DMF_FIXED_HEIGHT) height = Self->Height;
         else height = containerheight * Self->Height / 100;
         if (Self->Dimensions & DMF_FIXED_Y) value = containerheight - (Self->YCoord + height);
         else value = containerheight - ((Self->YCoord * containerheight / 100) + height);
      }
      else value = 0;

      if (Value->Type & FD_PERCENTAGE) value = (value * 100) / containerheight;
   }
   else return(ObjectError(Self, ERH_GetField, ERR_ExclusiveDenied));

   if (Value->Type & FD_DOUBLE) Value->Double = value;
   else if (Value->Type & FD_LARGE) Value->Large = value;
   else return(ObjectError(Self, ERH_GetField, ERR_FieldTypeMismatch));
   return(ERR_Okay);
}

static ERROR SET_YOffset(struct Arrow *Self, struct Variable *Value)
{
   if (Value->Type & FD_DOUBLE) Self->YOffset= Value->Double;
   else if (Value->Type & FD_LARGE) Self->YOffset= Value->Large;
   else return(ObjectError(Self, ERH_SetField, ERR_FieldTypeMismatch));

   if (Value->Type & FD_PERCENTAGE) Self->Dimensions = (Self->Dimensions & ~DMF_FIXED_YOFFSET) | DMF_RELATIVE_YOFFSET;
   else Self->Dimensions = (Self->Dimensions & ~DMF_RELATIVE_YOFFSET) | DMF_FIXED_YOFFSET;
   return(ERR_Okay);
}
