
static ERROR ResizeLayer(struct Render *, LONG, LONG, LONG, LONG, LONG, LONG, LONG, DOUBLE, LONG);

/*****************************************************************************

###ACTION###
Name:  Redimension
Short: Moves and resizes a render object in a single action call.
###END###

*****************************************************************************/

static ERROR RENDER_Redimension(struct Render *Self, struct acRedimension *Args)
{
   LONG oldx, oldy, oldwidth, oldheight;
   LONG newwidth, newheight, newx, newy, index;
   ERROR error;
   UBYTE msgbuffer[sizeof(struct Message) + sizeof(struct ActionMessage)];
   struct ActionMessage *action;
   APTR queue;

   if (!Args) return(ObjectError(Self, ERH_Redimension, ERR_Args)|ERF_Notified);

   /* Check if other redimension messages are queued for this object - if so,
   ** do not do anything until the final message is reached.
   */

   if (AccessMemory(GetResource(RES_MESSAGEQUEUE), MEM_READ, &queue) IS ERR_Okay) {
      index = 0;
      while (ScanMessages(queue, &index, MSGID_ACTION, msgbuffer, sizeof(msgbuffer)) IS ERR_Okay) {
         action = (struct ActionMessage *)(msgbuffer + sizeof(struct Message));
         if ((action->ActionID IS AC_Redimension) AND (action->ObjectID IS Self->Head.UniqueID)) {
            ReleaseMemory(queue);
            return(ERR_Okay|ERF_Notified);
         }
      }
      ReleaseMemory(queue);
   }

   #ifdef DBG_DRAW
      DPrintF("~Redimension()","[Render:%d] %dx%d,%dx%d $%.8x", Self->Head.UniqueID, (LONG)Args->XCoord, (LONG)Args->YCoord, (LONG)Args->Width, (LONG)Args->Height, Self->Flags);
   #endif

   if ((Args->Width < 0) OR (Args->Height < 0)) {
      #ifdef DBG_DRAW
         StepBack();
      #endif
      return(ERR_Args|ERF_Notified);
   }

   oldx      = Self->XCoord;
   oldy      = Self->YCoord;
   oldwidth  = Self->Width;
   oldheight = Self->Height;

   /*** Extract the new dimensions from the arguments ***/

   newx = Args->XCoord;
   newy = Args->YCoord;
   if (!Args->Width) newwidth = Self->Width;
   else newwidth = Args->Width;
   if (!Args->Height) newheight = Self->Height;
   else newheight = Args->Height;

   /*** Ensure that the requested width does not exceed minimum and maximum values ***/

   if ((Self->MinWidth > 0) AND (newwidth < Self->MinWidth + Self->LeftMargin + Self->RightMargin)) {
      if (oldwidth > newwidth) {
         if (oldwidth > Self->MinWidth + Self->LeftMargin + Self->RightMargin) newwidth = Self->MinWidth + Self->LeftMargin + Self->RightMargin;
         else newwidth = oldwidth; /* Maintain the current width because it is < MinWidth */
      }
   }

   if ((Self->MaxWidth > 0) AND (newwidth > Self->MaxWidth + Self->LeftMargin + Self->RightMargin)) newwidth = Self->MaxWidth + Self->LeftMargin + Self->RightMargin;

   if (newwidth < 2) newwidth = 2;

   /*** Check requested height against minimum and maximum height values ***/

   if ((Self->MinHeight > 0) AND (newheight < Self->MinHeight + Self->TopMargin + Self->BottomMargin)) {
      if (oldheight > newheight) {
         if (oldheight > Self->MinHeight + Self->TopMargin + Self->BottomMargin) newheight = Self->MinHeight + Self->TopMargin + Self->BottomMargin;
         else newheight = oldheight;
      }
   }

   if ((Self->MaxHeight > 0) AND (newheight > Self->MaxHeight + Self->TopMargin + Self->BottomMargin)) newheight = Self->MaxHeight + Self->TopMargin + Self->BottomMargin;

   if (newheight < 2) newheight = 2;

   /*** Check for changes ***/

   if ((newx IS oldx) AND (newy IS oldy) AND (newwidth IS oldwidth) AND (newheight IS oldheight)) {
      #ifdef DBG_DRAW
         StepBack();
      #endif
      return(ERR_Okay|ERF_Notified);
   }

   /*** Do the resize ***/

   error = ResizeLayer(Self, newx, newy, newwidth, newheight, newwidth, newheight, 0, 0.0, NULL);

   #ifdef DBG_DRAW
      StepBack();
   #endif
   return(error|ERF_Notified);
}

/*****************************************************************************

###ACTION###
Name:  Resize
Short: Resizes a render object's dimensions.
###END###

*****************************************************************************/

static ERROR RENDER_Resize(struct Render *Self, struct acResize *Args)
{
   struct acRedimension redimension;

   if (!Args) return(ObjectError(Self, ERH_Resize, ERR_Args|ERF_Notified));

   if (((!Args->Width) OR (Args->Width IS Self->Width)) AND
       ((!Args->Height) OR (Args->Height IS Self->Height))) return(ERR_Okay|ERF_Notified);

   redimension.XCoord = Self->XCoord;
   redimension.YCoord = Self->YCoord;
   redimension.ZCoord = 0;
   redimension.Width  = Args->Width;
   redimension.Height = Args->Height;
   redimension.Depth  = 0;
   return(Action(AC_Redimension, Self, &redimension)|ERF_Notified);
}

/*****************************************************************************

###METHOD###
Name:    SetDisplay
Synonym: SetRenderDisplay
Short:   Changes the screen resolution (applies to top-level render objects only).

###DESCRIPTION###
<p>The SetDisplay method is used to change the screen resolution of the top-level
render object (which represents the screen display).  It allows you to set the
size of the display and you may also change the bitmap depth and the monitor's
refresh rate.  If successful, the change is immediate.</p>

<p>This method exercises some intelligence in adjusting the display to your
requested settings.  For instance, if the requested width and/or height is not
available, the closest display setting will be chosen.</p>

<p>This method does not work on anything other than top-level render objects.
The current top-level render object is usually named "SystemDisplay" by default
and can be searched for by that name.</p>

###INPUT###
LONG/XCoord: The horizontal coordinate/offset for the display.
LONG/YCoord: The vertical coordinate/offset for the display.
LONG/Width: The width of the display.
LONG/Height: The height of the display.
LONG/InsideWidth: The page width of the display must be the same as Width or greater.
LONG/InsideHeight: The page height of the display must be the same as Height or greater.
LONG/BitsPerPixel: Bits per pixel - 15, 16, 24 or 32.
LONG/RefreshRate: Refresh rate.
LONG/Flags: Optional flags.

###END###

*****************************************************************************/

static ERROR RENDER_SetDisplay(struct Render *Self, struct mtSetRenderDisplay *Args)
{
   LONG newwidth, newheight, newx, newy;
   ERROR error;

   if ((!Args) OR (Args->Width < 0) OR (Args->Height < 0)) return(ObjectError(Self, MT_SetRenderDisplay, ERR_Args));
   if (Self->ParentID) return(ObjectError(Self, MT_SetRenderDisplay, ERR_Failed));

   DPrintF("~3SetDisplay()","[Render:%d] %dx%d,%dx%d", Self->Head.UniqueID, (LONG)Args->XCoord, (LONG)Args->YCoord, (LONG)Args->Width, (LONG)Args->Height);

   newx = Args->XCoord;
   newy = Args->YCoord;

   if (!Args->Width) newwidth = Self->Width;
   else newwidth = Args->Width;
   if (!Args->Height) newheight = Self->Height;
   else newheight = Args->Height;

   error = ResizeLayer(Self, newx, newy, newwidth, newheight,
      Args->InsideWidth, Args->InsideHeight, Args->BitsPerPixel,
      Args->RefreshRate, Args->Flags);

   StepBack();
   return(error);
}

/*****************************************************************************
** Internal: ResizeLayer()
**
** This function is responsible for managing the resizing of top-most render
** objects and is also used by some of the field management functions for
** Width/Height adjustments.
**
** This function is also useful for skipping the dimension limits normally
** imposed when resizing.
*/

static ERROR ResizeLayer(struct Render *Self, LONG XCoord, LONG YCoord,
                LONG Width, LONG Height, LONG InsideWidth, LONG InsideHeight,
                LONG BPP, DOUBLE RefreshRate, LONG DeviceFlags)
{
   struct RenderList *list;
   struct Bitmap *bitmap;
   struct acRedimension redimension;
   struct ClipRectangle old;
   WORD index;
   OBJECTPTR screen;

   if (!Width)  Width = Self->Width;
   if (!Height) Height = Self->Height;

   #ifdef DBG_DRAW
      DPrintF("~ResizeLayer()","[Render:%d:%s] %dx%d TO %dx%d", Self->Head.UniqueID, Self->Head.Stats->Name, Self->Width, Self->Height, Width, Height);
   #endif

   if (!(Self->Head.ObjectFlags & NF_INITIALISED)) {
      Self->XCoord = XCoord;
      Self->YCoord = YCoord;
      Self->Width  = Width;
      Self->Height = Height;
      #ifdef DBG_DRAW
         StepBack();
      #endif
      return(ERR_Okay);
   }

   if ((Self->XCoord IS XCoord) AND (Self->YCoord IS YCoord) AND
       (Self->Width IS Width) AND (Self->Height IS Height) AND
       (Self->ParentID)) {
      #ifdef DBG_DRAW
         StepBack();
      #endif
      return(ERR_Okay);
   }

   /* NB: It is absolutely vital that full-access is gained to the renderlist
   ** in order to prevent dead-locking between tasks during the resize process.
   ** (Dead-locking can occur if there is an intense graphics task or game
   ** running at the time).
   */

   if (!(list = AccessRenderList(ARF_WRITE, NULL))) {
      DPrintF("@ResizeLayer:","Unable to access the renderlist.");
      #ifdef DBG_DRAW
         StepBack();
      #endif
      return(ERR_AccessMemory);
   }

   index = FindDrawableIndex(list, Self->Head.UniqueID);

   if (Self->prvBitmapOwnerID IS Self->Head.UniqueID) {
      if (AccessObject(Self->BufferID, 5000, &bitmap) IS ERR_Okay) {
         if (ActionTags(AC_Resize, bitmap, (DOUBLE)Width, (DOUBLE)Height, 0.0) IS ERR_Okay) {
            Self->prvLineWidth     = bitmap->LineWidth;
            Self->prvBytesPerPixel = bitmap->BytesPerPixel;
            Self->prvBitsPerPixel  = bitmap->BitsPerPixel;
            Self->prvDataMID       = bitmap->DataMID;
            UpdateRenderList(Self);
         }
         else {
            ReleaseObject(bitmap);
            ReleaseRenderList(ARF_WRITE);
            return(StepObjectError(Self, ERH_Resize, ERR_Resize));
         }

         ReleaseObject(bitmap);
      }
      else {
         ReleaseRenderList(ARF_WRITE);
         return(StepObjectError(Self, ERH_Resize, ERR_AccessObject));
      }
   }

   if (!Self->ParentID) {
      struct mtSetScreenDisplay display;

      if (Width  > Self->MaxWidth  + Self->LeftMargin + Self->RightMargin)  Width  = Self->MaxWidth  + Self->LeftMargin + Self->RightMargin;
      if (Height > Self->MaxHeight + Self->TopMargin  + Self->BottomMargin) Height = Self->MaxHeight + Self->TopMargin  + Self->BottomMargin;
      if (InsideWidth < Width) InsideWidth = Width;
      if (InsideHeight < Height) InsideHeight = Height;


      if (AccessObject(Self->ScreenID, 5000, &screen) IS ERR_Okay) {
         display.XCoord       = XCoord;
         display.YCoord       = YCoord;
         display.Width        = Width;
         display.Height       = Height;
         display.InsideWidth  = InsideWidth;
         display.InsideHeight = InsideHeight;
         display.BitsPerPixel = BPP;
         display.RefreshRate  = RefreshRate;
         display.Flags        = DeviceFlags;
         if (Action(MT_SetScreenDisplay, screen, &display) != ERR_Okay) {
            ReleaseObject(screen);
            ReleaseRenderList(ARF_WRITE);
            return(StepObjectError(Self, ERH_Redimension, ERR_Redimension));
         }

         GetFields(screen, FID_Width|TLONG, &Width, FID_Height|TLONG, &Height, TAGEND);

         ReleaseObject(screen);
      }
      else {
         ReleaseRenderList(ARF_WRITE);
         return(StepObjectError(Self, ERH_Redimension, ERR_AccessObject));
      }
   }

   old.left   = list[index].Left;
   old.top    = list[index].Top;
   old.right  = list[index].Right;
   old.bottom = list[index].Bottom;

   Self->XCoord = XCoord;
   Self->YCoord = YCoord;
   Self->Width  = Width;
   Self->Height = Height;
   UpdateRenderList(Self);

   if (!(Self->Head.ObjectFlags & NF_INITIALISED)) {
      ReleaseRenderList(ARF_WRITE);
      #ifdef DBG_DRAW
         StepBack();
      #endif
      return(ERR_Okay);
   }

   /* Broadcast a Resize notification to our subscribers.  Basically, this
   ** informs our drawable children to resize themselves to the new dimensions.
   ** Drawable objects are not permitted to redraw themselves when they receive
   ** the Redimension notification - we will send a delayed draw message later
   ** in this routine.
   */

   glNoDrawing++;

   redimension.XCoord = Self->XCoord;
   redimension.YCoord = Self->YCoord;
   redimension.ZCoord = 0;
   redimension.Width  = Self->Width;
   redimension.Height = Self->Height;
   redimension.Depth  = 0;
   NotifySubscribers(Self, AC_Redimension, &redimension, NULL);

   glNoDrawing--;

   if (!(Self->Flags & RNF_VISIBLE)) {
      ReleaseRenderList(ARF_WRITE);
      #ifdef DBG_DRAW
         StepBack();
      #endif
      return(ERR_Okay);
   }

   if (!glNoDrawing) {
      WORD vindex;

      /* Post the drawing update using RedrawDrawable().  This is the only
      ** reliable way to generate updates when our drawable may contain
      ** children that belong to foreign tasks.
      */

      #ifdef DBG_DRAW
         DPrintF("ResizeLayer:","Redrawing the resized drawable.");
      #endif

      RedrawDrawable(Self->Head.UniqueID, list[index].AbsX, list[index].AbsY, list[index].Right, list[index].Bottom, NULL);
      ExposeDrawable(Self->Head.UniqueID, 0, 0, Self->Width, Self->Height, EXF_CHILDREN|EXF_REDRAW_VOLATILE_OVERLAP);

      /* Update external regions on all four sides that have been exposed
      ** by the resize, for example due to a decrease in area or a coordinate
      ** shift.
      **
      ** Note: glVolatileIndex determines the point at which volatile exposes will start.
      ** We want volatile exposes to start just after our target drawable, and not
      ** anything that sits behind us in the containing parent.
      */

      for (vindex=index+1; list[vindex].Level > list[index].Level; vindex++);
      glVolatileIndex = vindex;

      if (Self->prvBitmapOwnerID IS Self->Head.UniqueID) {
         RedrawNonIntersect(Self->ParentID, (struct ClipRectangle *)(&list[index].AbsX), &old, -1, EXF_CHILDREN|EXF_REDRAW_VOLATILE);
      }
      else RedrawNonIntersect(Self->ParentID, (struct ClipRectangle *)(&list[index].AbsX), &old, NULL, EXF_CHILDREN|EXF_REDRAW_VOLATILE);

      glVolatileIndex = 0;
   }

   ReleaseRenderList(ARF_WRITE);

   #ifdef DBG_DRAW
      StepBack();
   #endif
   return(ERR_Okay);
}
